#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <sched.h>
#include <time.h>

#include "config.h"

typedef struct {
	unsigned int id;
	pthread_t thread;
	bool sense;
} thread_t;

bool sense;
unsigned int count;

static void dobarrier (thread_t * const t) __attribute__((always_inline));

static void dobarrier (thread_t * const t) {
	t->sense = !t->sense;
	/* atomic decrement */
	if (__sync_fetch_and_sub (&count, 1) == 1) {
		/* last processor */
		count = THREAD_NUM;
		sense = t->sense;
	} else {
		while (sense != t->sense) {
			sched_yield ();
		}
	}
}

void *worker (void *arg) {
	thread_t *info = arg;
	const size_t runs = (info->id == 0) ? UNBALANCEDADDRUNS : ADDRUNS;

	for (size_t i = 0; i < BARRIERRUNS; i++) {
		volatile unsigned int foo = 0;
		for (size_t j = 0; j < runs; j++) {
			foo += j;
		}
		debug ("worker %u done, %u\n", info->id, foo);
		dobarrier (info);
	}
	return NULL;
}

int main () {
	thread_t w[THREAD_NUM];
	struct timespec start, end;

	sense = false;
	count = THREAD_NUM;

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		w[i].id = i;
		w[i].sense = false;
	}

	clock_gettime (CLOCK_MONOTONIC_RAW, &start);
	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_create (&w[i].thread, NULL, worker, &w[i]);
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_join (w[i].thread, NULL);
	}
	clock_gettime (CLOCK_MONOTONIC_RAW, &end);

	unsigned long long int t = (end.tv_sec - start.tv_sec) * 1000000000ULL +
			(end.tv_nsec - start.tv_nsec);
	printf ("%llu\n", t);
}

