#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <sched.h>

#include "config.h"

typedef struct {
	bool myflags[2][THREAD_NUM_LOG2];
	bool *partnerflags[2][THREAD_NUM_LOG2];

	int parity;
	bool sense;
} barriernode;

typedef struct {
	unsigned int id;
	pthread_t thread;
} threadinfo;

barriernode nodes[THREAD_NUM];

void dobarrier (unsigned int id) {
	barriernode * const n = &nodes[id];
	for (size_t i = 0; i < THREAD_NUM_LOG2; i++) {
		//printf ("thread %i sets partner[%u][%u] to %u\n", id, n->parity, i, n->sense);
		*n->partnerflags[n->parity][i] = n->sense;
		while (n->myflags[n->parity][i] != n->sense) {
			sched_yield ();
		}
	}
	if (n->parity == 1) {
		n->sense = !n->sense;
	}
	n->parity = 1 - n->parity;
}

void *worker (void *arg) {
	threadinfo *info = arg;

	for (size_t i = 0; i < BARRIERRUNS; i++) {
		unsigned int foo = 0;
		for (size_t j = 0; j < ADDRUNS; j++) {
			foo += j;
		}
		printf ("worker %u done, %u\n", info->id, foo);
		dobarrier (info->id);
	}
	return NULL;
}

unsigned int naivepow2 (unsigned int k) {
	unsigned int ret = 1;
	while (k > 0) {
		ret *= 2;
		--k;
	}
	return ret;
}

int main () {
	threadinfo w[THREAD_NUM];

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		w[i].id = i;

		/* init barrier */
		printf ("init for %u\n", i);
		memset (&nodes[i], 0, sizeof (nodes[i]));
		nodes[i].sense = true;
		for (size_t r = 0; r < 2; r++) {
			for (size_t k = 0; k < THREAD_NUM_LOG2; k++) {
				unsigned int j = (i+naivepow2(k)) % THREAD_NUM;
				printf ("%u, %u, %u -> %u, %u, %u\n", i, r, k, j, r, k);
				nodes[i].partnerflags[r][k] =
						&nodes[j].myflags[r][k];
			}
		}
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_create (&w[i].thread, NULL, worker, &w[i]);
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_join (w[i].thread, NULL);
	}
}

