#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <sched.h>
#include <time.h>

#include "config.h"

typedef struct {
	bool mine;
	bool *partner;
} flags_t;

typedef struct {
	flags_t flags[2][THREAD_NUM_LOG2];

	bool sense;
	unsigned int parity;
} barriernode;

typedef struct {
	unsigned int id;
	pthread_t thread;
	barriernode *n;
} threadinfo;

static void dobarrier (barriernode * const n) __attribute__((always_inline));

static void dobarrier (barriernode * const n) {
	for (unsigned int i = 0; i < THREAD_NUM_LOG2; i++) {
		flags_t * const flags = &n->flags[n->parity][i];
		*flags->partner = n->sense;
		while (flags->mine != n->sense) {
			sched_yield ();
		}
	}
	if (n->parity == 1) {
		n->sense = !n->sense;
	}
	n->parity ^= 0x1;
}

void *worker (void *arg) {
	threadinfo *info = arg;
	const size_t runs = (info->id == 0) ? UNBALANCEDADDRUNS : ADDRUNS;

	for (size_t i = 0; i < BARRIERRUNS; i++) {
		volatile unsigned int foo = 0;
		for (size_t j = 0; j < runs; j++) {
			foo += j;
		}
		debug ("worker %u done, %u\n", info->id, foo);
		dobarrier (info->n);
	}
	return NULL;
}

/*	naive, constant-space power of 2
 */
unsigned int naivepow2 (unsigned int k) {
	unsigned int ret = 1;
	while (k > 0) {
		ret *= 2;
		--k;
	}
	return ret;
}

int main () {
	threadinfo w[THREAD_NUM];
	barriernode nodes[THREAD_NUM];
	struct timespec start, end;

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		w[i].id = i;
		w[i].n = &nodes[i];

		/* init barrier */
		debug ("init for %u\n", i);
		memset (&nodes[i], 0, sizeof (nodes[i]));
		nodes[i].sense = true;
		for (size_t r = 0; r < 2; r++) {
			for (size_t k = 0; k < THREAD_NUM_LOG2; k++) {
				unsigned int j = (i+naivepow2(k)) % THREAD_NUM;
				debug ("%u, %u, %u -> %u, %u, %u\n", i, r, k, j, r, k);
				nodes[i].flags[r][k].partner =
						&nodes[j].flags[r][k].mine;
			}
		}
	}

	clock_gettime (CLOCK_MONOTONIC_RAW, &start);
	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_create (&w[i].thread, NULL, worker, &w[i]);
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_join (w[i].thread, NULL);
	}
	clock_gettime (CLOCK_MONOTONIC_RAW, &end);

	unsigned long long int t = (end.tv_sec - start.tv_sec) * 1000000000ULL +
			(end.tv_nsec - start.tv_nsec);
	printf ("%llu\n", t);
}

