#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include <linux/futex.h>

#include "config.h"

typedef struct {
	uint8_t parentsense;
	uint8_t *parent;
	uint8_t *children[2];
	union {
		uint8_t single[4];
		uint32_t all;
	} havechild;
	union {
		uint8_t single[4];
		uint32_t all;
	} childnotready;
	int childnotreadyfutex;
	int *parentnotify;
	uint8_t sense;
	uint8_t dummy;
	int dummywide;
} barriernode;

typedef struct {
	unsigned int id;
	pthread_t thread;
} thread_t;

barriernode nodes[THREAD_NUM];

#include <unistd.h>
#include <sys/syscall.h>

int sys_futex(int *uaddr, int op, int val, const struct timespec * timeout) {
	return syscall (SYS_futex, uaddr, op, val, timeout);
}

void dobarrier (unsigned int id) {
	barriernode * const n = &nodes[id];
	/* spin */
	while (1) {
		volatile uint32_t val = n->childnotready.all;
		if (val == 0) {
			break;
		}
		sys_futex (&n->childnotreadyfutex, FUTEX_WAIT, val, NULL);
	}
	n->childnotready.all = n->havechild.all;
	/* byte stores are atomic on amd64 */
	*n->parent = 0x0;
	sys_futex (n->parentnotify, FUTEX_WAKE, 1000, NULL);
	if (id != 0) {
		/* spin */
		while (n->parentsense != n->sense) {
			sched_yield ();
		}
	}
	*n->children[0] = n->sense;
	*n->children[1] = n->sense;
	n->sense = !n->sense;
}

void *worker (void *arg) {
	thread_t *info = arg;

	for (size_t i = 0; i < BARRIERRUNS; i++) {
		unsigned int foo = 0;
		for (size_t j = 0; j < ADDRUNS; j++) {
			foo += j;
		}
#if 0
		printf ("worker %u done, %u\n", info->id, foo);
#endif
		dobarrier (info->id);
	}
	printf ("worker %u done\n", info->id);
	return NULL;
}

int main () {
	thread_t w[THREAD_NUM];

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		w[i].id = i;

		/* init barrier */
		printf ("init for %u\n", i);
		for (size_t j = 0; j < 4; j++) {
			nodes[i].havechild.single[j] = (4*i+j < (THREAD_NUM-1)) ? true : false;
			printf ("havechild[%u]=%i\n", j, nodes[i].havechild.single[j]);
		}
		nodes[i].parent = (i != 0) ? &nodes[(i-1)/4].childnotready.single[(i-1)%4] : &nodes[i].dummy;
		nodes[i].parentnotify = (i != 0) ? &nodes[(i-1)/4].childnotreadyfutex : &nodes[i].dummywide;
		nodes[i].childnotreadyfutex = 0;
		if (i != 0) {
			printf ("parent = nodes[%u].childnotready[%u]\n", (i-1)/4, (i-1)%4);
		}
		nodes[i].children[0] = (2*i+1 < THREAD_NUM) ? &nodes[2*i+1].parentsense : &nodes[i].dummy;
		if (2*i+1 < THREAD_NUM) {
			printf ("children[0] = nodes[%u]\n", 2*i+1);
		}
		nodes[i].children[1] = (2*i+2 < THREAD_NUM) ? &nodes[2*i+2].parentsense : &nodes[i].dummy;
		if (2*i+2 < THREAD_NUM) {
			printf ("children[1] = nodes[%u]\n", 2*i+2);
		}
		nodes[i].childnotready.all = nodes[i].havechild.all;
		nodes[i].parentsense = 0x0;
		nodes[i].sense = 0x1;
		nodes[i].dummy = 0x0;
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_create (&w[i].thread, NULL, worker, &w[i]);
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_join (w[i].thread, NULL);
	}
}

