#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>

#include "config.h"

typedef struct {
	bool parentsense;
	bool *parent;
	bool *children[2];
	bool havechild[4];
	bool childnotready[4];
	bool sense;
	bool dummy;
} barriernode;

typedef struct {
	unsigned int id;
	pthread_t thread;
} threadinfo;

barriernode nodes[THREAD_NUM];

void dobarrier (unsigned int id) {
	barriernode * const n = &nodes[id];
	/* spin */
	while (n->childnotready[0] || n->childnotready[1] || n->childnotready[2] || n->childnotready[3]);
	memcpy (n->childnotready, n->havechild, sizeof (n->childnotready));
	*n->parent = false;
	if (id != 0) {
		while (n->parentsense != n->sense);
	}
	*n->children[0] = n->sense;
	*n->children[1] = n->sense;
	n->sense = !n->sense;
}

void *worker (void *arg) {
	threadinfo *info = arg;

	for (size_t i = 0; i < BARRIERRUNS; i++) {
		unsigned int foo = 0;
		for (size_t j = 0; j < ADDRUNS; j++) {
			foo += j;
		}
		//printf ("worker %u done, %u\n", info->id, foo);
		dobarrier (info->id);
	}
	return NULL;
}

int main () {
	threadinfo w[THREAD_NUM];

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		w[i].id = i;

		/* init barrier */
		printf ("init for %u\n", i);
		for (size_t j = 0; j < 4; j++) {
			nodes[i].havechild[j] = (4*i+j < (THREAD_NUM-1)) ? true : false;
			printf ("havechild[%u]=%i\n", j, nodes[i].havechild[j]);
		}
		nodes[i].parent = (i != 0) ? &nodes[(i-1)/4].childnotready[(i-1)%4] : &nodes[i].dummy;
		if (i != 0) {
			printf ("parent = nodes[%u].childnotready[%u]\n", (i-1)/4, (i-1)%4);
		}
		nodes[i].children[0] = (2*i+1 < THREAD_NUM) ? &nodes[2*i+1].parentsense : &nodes[i].dummy;
		if (2*i+1 < THREAD_NUM) {
			printf ("children[0] = nodes[%u]\n", 2*i+1);
		}
		nodes[i].children[1] = (2*i+2 < THREAD_NUM) ? &nodes[2*i+2].parentsense : &nodes[i].dummy;
		if (2*i+2 < THREAD_NUM) {
			printf ("children[1] = nodes[%u]\n", 2*i+2);
		}
		memcpy (nodes[i].childnotready, nodes[i].havechild, sizeof (nodes[i].childnotready));
		nodes[i].parentsense = false;
		nodes[i].sense = true;
		nodes[i].dummy = false;
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_create (&w[i].thread, NULL, worker, &w[i]);
	}

	for (size_t i = 0; i < sizeof (w)/sizeof (*w); i++) {
		pthread_join (w[i].thread, NULL);
	}
}

